import pytest
import numpy as np
import pandas as pd

from inspec.preprocessing.predictive_variables import get_predictive_features


def test__get_predictive_features__when_single_columns_has_positive_coefficient__works():
    x_train = np.array(
        [
            [1, 3],
            [2, 1],
            [3, 4],
            [4, 1],
            [5, 5],
            [6, 9],
            [7, 2],
            [8, 6],
            [9, 5],
            [10, 3],
        ]
    )

    y_train = x_train[:, 0]

    optimal_features, coefficients = get_predictive_features(x_train, y_train)

    assert len(x_train) == len(optimal_features)
    assert 1 == optimal_features.shape[1]
    assert (np.reshape(x_train[:, 0], (10, 1)) == optimal_features).all()

    assert coefficients[0] > 0
    assert coefficients[1] == 0


def test__get_predictive_features__when_multiple_columns_have_positive_coefficient__works():
    x_train = np.array(
        [
            [1, 11, 3],
            [2, 12, 1],
            [3, 13, 4],
            [4, 14, 1],
            [5, 15, 5],
            [6, 16, 9],
            [7, 17, 2],
            [8, 18, 6],
            [9, 19, 5],
            [10, 20, 3],
        ]
    )

    y_train = x_train[:, 0] + x_train[:, 1]

    optimal_features, coefficients = get_predictive_features(x_train, y_train)

    assert len(x_train) == len(optimal_features)
    assert 2 == optimal_features.shape[1]
    assert (np.reshape(x_train[:, 0:2], (10, 2)) == optimal_features).all()

    assert coefficients[0] > 0
    assert coefficients[1] > 0
    assert coefficients[2] <= 0


def test__get_predictive_features__when_predictions_all_zeros__returns_empty():
    x_train = np.array(
        [
            [1, 3],
            [2, 1],
            [3, 4],
            [4, 1],
            [5, 5],
            [6, 9],
            [7, 2],
            [8, 6],
            [9, 5],
            [10, 3],
        ]
    )

    y_train = np.zeros(shape=(len(x_train),))

    optimal_features, coefficients = get_predictive_features(x_train, y_train)

    assert 0 == optimal_features.size

    assert (coefficients == 0).all()


def test__get_predictive_features__when_pandas__works():
    x_train = np.array(
        [
            [1, 3],
            [2, 1],
            [3, 4],
            [4, 1],
            [5, 5],
            [6, 9],
            [7, 2],
            [8, 6],
            [9, 5],
            [10, 3],
        ]
    )

    y_train = x_train[:, 0]

    optimal_features, coefficients = get_predictive_features(
        pd.DataFrame(data=x_train), pd.DataFrame(data=y_train)
    )

    assert len(x_train) == len(optimal_features)
    assert 1 == optimal_features.shape[1]
    assert (np.reshape(x_train[:, 0], (10, 1)) == optimal_features).all()

    assert coefficients[0] > 0
    assert coefficients[1] == 0


def test__get_predictive_features__when_inputs_do_not_have_same_size__raises():
    x_train = np.array([[1, 3], [2, 1], [3, 4], [4, 1]])

    y_train = np.array([1, 2])

    with pytest.raises(ValueError) as e:
        _, _ = get_predictive_features(x_train, y_train)

    assert (
        str(e.value)
        == "The provided inputs must contain the same number of elements: x_train and y_train."
    )



from typing import Tuple, Union

import numpy as np
import pandas as pd

from sklearn.linear_model import Lasso


def get_predictive_features(
    x_train: Union[pd.DataFrame, np.ndarray],
    y_train: Union[pd.DataFrame, np.ndarray],
    alpha: float = 0.1,
) -> Tuple[np.ndarray, np.ndarray]:
    """Evaluates the optimal features of a dataset based on the coefficients returned by a Lasso model used as a baseline.

    Args:
        x_train: Numpy array or pandas Series. Training dataset.
        y_train: Numpy array or pandas Series. Predicted target values.

    Returns: 
        The dataset with only the optimal features.
        The list of coefficients used to establish the optimal features

    """

    _x_train = _pandas_to_ndarray(x_train)
    _y_train = _pandas_to_ndarray(y_train)

    _validate_inputs(_x_train, _y_train)

    model = Lasso(alpha=alpha)
    model.fit(_x_train, _y_train)

    optimal_features = _x_train[:, (model.coef_ != 0)]

    return optimal_features, model.coef_


def _pandas_to_ndarray(data):
    if isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
        return data.values

    if isinstance(data, np.ndarray):
        return data

    if isinstance(data, list):
        return np.array(data)

    raise ValueError(
        f"Inputs must be Pandas Dataframe or Numpy arrays. Cannot convert {type(data)} to numpy array."
    )


def _validate_inputs(x_train: np.ndarray, y_train: np.ndarray) -> None:
    if len(x_train) != len(y_train):
        raise ValueError(
            "The provided inputs must contain the same number of elements: x_train and y_train."
        )




import copy
import warnings

import numpy as np
import pandas as pd

from inspec.utils.text_utils import str_as_list


def mean_absolute_scaled_error(y_true, y_pred) -> float:
    """Computes the mean absolute scaled error.

    Args:
        y_true: Numpy array or pandas Series. Ground truth (correct) target values.
        y_pred: Numpy array or pandas Series. Predicted target values.

    Returns: The mean absolute scaled error.

    """
    inputs_dict = { "y_true": y_true, "y_pred": y_pred }

    _validate_input_types(inputs_dict)
    _validate_input_sizes(inputs_dict)

    _y_true = _pandas_to_ndarray(y_true)
    _y_pred = _pandas_to_ndarray(y_pred)

    absolute_residual = np.abs(_y_true - _y_pred)
    naive_prediction = _y_true[:-1]
    scale = _mean_absolute_error(_y_true[1:], naive_prediction)
    scale += 0.0000000000000001  # to avoid dividing by zero

    return float(np.mean(absolute_residual / scale))


def mase_by_series(y_true, y_pred, dimension=None, time=None) -> pd.Series:
    """Computes the mean absolute scaled error by series.

    Works for a single time series and multiple time series datasets (eg. multiple products through time).

    Assumes that there is only one dimension when none is provided as an input.

    Assumes that the observations are already sorted in chronological order, when no time is provided.

    Args:
        y_true: Numpy array or pandas Series. Ground truth (correct) target values.
        y_pred: Numpy array or pandas Series. Predicted target values.
        dimension: Numpy array or pandas Series. Column with identifiers to group by. I could be products, for example.
        time: Numpy array or pandas Series. Column with time indices that can be sorted in increasing order.

    Returns: A series containing the mase for each dimension.

    """
    _dimension = copy.deepcopy(dimension)
    _time = copy.deepcopy(time)

    if _dimension is None:
        _dimension = np.full((len(y_true), 1), 0)
        warnings.warn(
            "No dimension specified; assuming that there is only one dimension."
        )

    if _time is None:
        _time = np.arange(len(y_true))
        warnings.warn(
            "No time index specified; using 'y_true.index'."
        )

    inputs_dict = {"y_true": y_true, "y_pred": y_pred, "dimension": _dimension, "time": _time}

    _validate_input_types(inputs_dict)
    _validate_input_sizes(inputs_dict)
    _validate_time_input(_time)

    panel = pd.DataFrame(
        inputs_dict
    )

    if panel.groupby("dimension").count()["y_true"].min() < 2:
        raise ValueError("All dimensions must contain more than one observation.")

    panel.sort_values(["dimension", "time"], inplace=True)

    mases = panel.groupby("dimension").apply(
        lambda x: mean_absolute_scaled_error(x["y_true"], x["y_pred"])
    )

    mases.sort_values(ascending=False, inplace=True)

    return mases


def _mean_absolute_error(y_true, y_pred) -> float:
    """Computes the mean absolute error.

    Args:
        y_true: Numpy array or pandas Series. Ground truth (correct) target values.
        y_pred: Numpy array or pandas Series. Predicted target values.

    Returns: The mean absolute error.

    """

    _validate_input_sizes({"y_true": y_true, "y_pred": y_pred})

    absolute_residual = np.abs(_pandas_to_ndarray(y_true) - _pandas_to_ndarray(y_pred))

    return absolute_residual.mean()


def _pandas_to_ndarray(data):
    if isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
        return data.values

    if isinstance(data, np.ndarray):
        return data

    raise ValueError(f"Cannot convert {type(data)} to numpy array.")


def _validate_input_types(inputs):
    for input_name, input in inputs.items():
        if type(input) not in [np.ndarray, pd.Series]:
            raise ValueError(
                f"{input_name} must be either a numpy array or a pandas Series."
            )


def _validate_input_sizes(inputs):
    input_lengths = [len(x) for x in inputs.values()]

    if input_lengths[0] < 2:
        raise ValueError(f"The provided inputs must contain at least 2 elements: {str_as_list(list(inputs))}.")

    if not all(x == input_lengths[0] for x in input_lengths):
        raise ValueError(f"The provided inputs must contain the same number of elements: {str_as_list(list(inputs))}.")


def _validate_time_input(time):
    if len(time.shape) > 1:
        raise ValueError("The time column must be one-dimensional.")

    if (
        "int" not in str(time.dtype)
        and "float" not in str(time.dtype)
        and "date" not in str(time.dtype)
    ):
        raise ValueError("The time column must contain int, float or dates.")

    if not isinstance(time, pd.Series) and not isinstance(time, np.ndarray):
        raise ValueError(
            "The time column must be either a numpy array or a pandas Series."
        )



import numpy as np
import pandas as pd
import pytest

from inspec.metrics.mase import (
    mase_by_series,
    mean_absolute_scaled_error,
    _mean_absolute_error,
)


@pytest.mark.parametrize(
    "y_true, y_pred, expected_output",
    [
        (np.array([0, 1, 2, 3]), np.array([1, 2, 3, 4]), 1.0),
        (np.array([0, 1, 2, 3]), np.array([0, 1, 2, 3]), 0.0),
        (np.array([0, 10, 20, 30]), np.array([10, 20, 30, 40]), 10.0),
        (np.array([0, 10, 20, 30]), np.array([-10, 20, -30, 40]), 20.0),
        (pd.Series([0, 1, 2, 3]), pd.Series([1, 2, 3, 4]), 1.0),
        (np.array([0, 1, 2, 3]), pd.Series([1, 2, 3, 4]), 1.0),
    ],
)
def test__mean_absolute_error__returns_correct_output(y_true, y_pred, expected_output):
    assert _mean_absolute_error(y_true, y_pred) == expected_output


@pytest.mark.parametrize(
    "y_true, y_pred, expected_output",
    [
        (np.array([0, 1, 2, 3]), np.array([1, 2, 3, 4]), 1.0),
        (np.array([0, 1, 2, 3]), np.array([0, 1, 2, 3]), 0.0),
        (np.array([0, 10, 20, 30]), np.array([10, 20, 30, 40]), 1.0),
        (pd.Series([0, 1, 2, 3]), pd.Series([1, 2, 3, 4]), 1.0),
        (np.array([0, 1, 2, 3]), pd.Series([1, 2, 3, 4]), 1.0),
    ],
)
def test__mean_absolute_scaled_error__returns_correct_output(
    y_true, y_pred, expected_output
):
    assert mean_absolute_scaled_error(y_true, y_pred) == expected_output


def test__mean_absolute_scaled_error__raise_when_y_true_has_different_size_than_y_pred():
    y_true = np.array([0, 1, 2, 3, 0, 1, 2, 3, 0, 10, 20])
    y_pred = np.array([1, 2, 3, 4, 0, 1, 2, 3, 10, 20, 30, 40])

    with pytest.raises(ValueError) as e:
        mean_absolute_scaled_error(y_true, y_pred)

    assert str(e.value) == "The provided inputs must contain the same number of elements: y_true and y_pred."


def test__mean_absolute_scaled_error__raise_when_y_true_has_less_than_two_values():
    y_true = np.array([0])
    y_pred = np.array([1])

    with pytest.raises(ValueError) as e:
        mean_absolute_scaled_error(y_true, y_pred)
        
    assert str(e.value) == "The provided inputs must contain at least 2 elements: y_true and y_pred."


def test__mase_by_series__returns_correct_output():
    y_true = np.array([0, 1, 2, 3, 0, 1, 2, 3, 0, 10, 20, 30])
    y_pred = np.array([1, 2, 3, 4, 0, 1, 2, 3, 10, 20, 30, 40])
    dimension = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])
    time = np.array([1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4])

    mases = mase_by_series(y_true, y_pred, dimension, time)

    expected_output = pd.Series({0: 1, 2: 1, 1: 0})

    assert isinstance(mases, pd.Series)
    assert (mases.index == expected_output.index).all()
    assert (mases.values == expected_output.values).all()


def test__mase_by_series__str_dimension__returns_correct_output():
    y_true = np.array([0, 1, 2, 3, 0, 1, 2, 3, 0, 10, 20, 30])
    y_pred = np.array([1, 2, 3, 4, 0, 1, 2, 3, 10, 20, 30, 40])
    dimension = np.array(["0", "0", "0", "0", "1", "1", "1", "1", "2", "2", "2", "2"])
    time = np.array([1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4])

    mases = mase_by_series(y_true, y_pred, dimension, time)

    expected_output = pd.Series({"0": 1, "2": 1, "1": 0})

    assert isinstance(mases, pd.Series)
    assert (mases.index == expected_output.index).all()
    assert (mases.values == expected_output.values).all()


def test__mase_by_series__works_with_date_dtype():
    y_true = np.array([0, 1, 2, 3, 0, 1, 2, 3, 0, 10, 20, 30])
    y_pred = np.array([1, 2, 3, 4, 0, 1, 2, 3, 10, 20, 30, 40])
    dimension = np.array(["0", "0", "0", "0", "1", "1", "1", "1", "2", "2", "2", "2"])
    time = np.array([np.datetime64(i, "Y") for i in range(0, 4)] * 3)

    mases = mase_by_series(y_true, y_pred, dimension, time)

    expected_output = pd.Series({"0": 1, "2": 1, "1": 0})

    assert isinstance(mases, pd.Series)
    assert (mases.index == expected_output.index).all()
    assert (mases.values == expected_output.values).all()


def test__mase_by_series__raises_when_dimension_not_same_size_as_y_true():
    y_true = np.array([0, 1, 0, 1, 0, 10])
    y_pred = np.array([1, 2, 0, 1, 10, 20])
    dimension = np.array(["0", "0", "0", "1", "1", "1", "2", "2", "2"])
    time = np.array([1, 2, 1, 2, 1, 2])

    with pytest.raises(ValueError) as e:
        mase_by_series(y_true, y_pred, dimension, time)
        
    assert str(e.value) == "The provided inputs must contain the same number of elements: y_true, y_pred, dimension and time."


def test__mase_by_series__raises_when_time_is_str():
    y_true = np.array([0, 1, 2, 3, 0, 1, 2, 3, 0, 10, 20, 30])
    y_pred = np.array([1, 2, 3, 4, 0, 1, 2, 3, 10, 20, 30, 40])
    dimension = np.array(["0", "0", "0", "0", "1", "1", "1", "1", "2", "2", "2", "2"])
    time = np.array(["1", "2", "3", "4", "1", "2", "3", "4", "1", "2", "3", "4"])

    with pytest.raises(ValueError) as e:
        mase_by_series(y_true, y_pred, dimension, time)
        
    assert str(e.value) == "The time column must contain int, float or dates."


def test__mase_by_series__raises_when_dimension_size_is_lower_than_two():
    y_true = np.array([0, 0, 1, 2, 3, 0, 10, 20, 30])
    y_pred = np.array([1, 0, 1, 2, 3, 10, 20, 30, 40])
    dimension = np.array(["0", "1", "1", "1", "1", "2", "2", "2", "2"])
    time = np.array([1, 1, 2, 3, 4, 1, 2, 3, 4])

    with pytest.raises(ValueError) as e:
        mase_by_series(y_true, y_pred, dimension, time)
        
    assert str(e.value) == "All dimensions must contain more than one observation."



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name = "inspec-ai",
    version = "0.1.2",
    license="Internal Use", # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description="Library containing all the prototypes that were developped by the Moov AI product team.",
    long_description=README,
    long_description_content_type="text/markdown",
    url = "https://moov.ai/en/",
    author = "MoovAI Technologies Inc",
    author_email = "info@moov.ai",
    classifiers=[
        "Development Status :: 3 - Alpha", # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9"
    ],
    packages = find_packages(),
    include_package_data=True,
    install_requires=[
        "fuzzywuzzy>=0.18",
        "numpy>=1",
        "pandas>=1",
        "pandas-profiling>=3",
        "scikit-learn>=1"
    ],
    entry_points={},
)


import copy

import pandas as pd
import pandas_profiling
import json
import os
import string
from fuzzywuzzy import fuzz
import itertools


class CategoryCleaner:
    def __init__(self, df: pd.DataFrame, pandas_profiling_report: dict=None):
        """

        :param df: pd.DataFrame containing the data to be cleaned. Assumes that df has column headers
        :param pandas_profiling_report: [OPTIONAL] the pandas profiling report associated with the df. If not provided, will be generated.
        """
        self.df = df

        if pandas_profiling_report:
            self.pandas_profiling_report = pandas_profiling_report
        else:
            self.pandas_profiling_report = self.get_pandas_profiling_report()


    def get_pandas_profiling_report(self) -> dict:
        """
        Generates a pandas profiling report for the pd.DataFrame given as input to the class

        :return: dict, containing the report's results
        """
        profile = pandas_profiling.ProfileReport(self.df, title="Pandas Profiling Report", explorative=True)
        return json.loads(profile.to_json())


    def get_string_categorical_features(self) -> list:
        """
        Finds the features that are categorical and of dtypes "object" to find string categorical features

        :return: list, containing strings. Each value is the name of a feature identified as categorical with string values.
        """
        # get categorical features
        categorical_features = []
        for feat in self.pandas_profiling_report["variables"]:
            if self.pandas_profiling_report["variables"][feat]["type"] == 'Categorical':
                categorical_features.append(feat)

        # get features with only strings (object) type
        types = self.df.dtypes
        object_type = list(types[types == "object"].index)

        string_categorical_features=[]
        # clean each string categorical feature
        for feat in categorical_features:
            if feat in object_type:
                string_categorical_features.append(feat)

        return string_categorical_features


    def get_unique_values(self, col:pd.Series) -> pd.Series:
        """
        Finds unique values contained in "col". Ignores NaN and empty strings as unique values

        :param col: pd.Series. The data to check the unique values in. It is assumed that col contains strings.
        :return: pd.Series. Contains the unique values found in col
        """
        values = pd.Series(col.dropna().unique())
        values = values[values != ""]
        return values


    def get_cleaned_values(self, unique_values:pd.Series) -> pd.Series:
        """
        Performs some cleaning operations on values in "unique_values". Assumes that "unique_values" contains strings.

        :param unique_values: pd.Series. Usually a column with the unique values of another column. Assumed to be strings.
        :return: pd.Series. The cleaned version of "unique_values"
        """
        # BASIC PRE-PROCESSING OF THE UNIQUE VALUES
        # remove whitespaces at the end and beginning
        clean_values = unique_values.str.strip()

        # replace multiple white spaces with single whitespace
        clean_values = clean_values.str.replace(r" +", r" ", regex=True)

        # only lower case
        clean_values = clean_values.str.lower()

        # remove punctuation
        clean_values = clean_values.str.replace(r"[{}]".format(string.punctuation), r"", regex=True)

        # remove digits
        clean_values = clean_values.str.replace(r"\d+", r"", regex=True)

        return clean_values


    def get_basic_groupings(self, col:pd.Series) -> list:
        """
        Gets groupings of values contained in "col" that can be easily identified as similar.

        :param col: pd.Series. A single column (feature). Assumed to contain strings.
        :return: list. Each element is a tuple of two strings. Each tuple represents the name of columns that could be grouped.
        """
        unique_values = self.get_unique_values(col)
        clean_unique_values = self.get_cleaned_values(unique_values)

        values_count = clean_unique_values.value_counts()

        # find the ones occuring more than once
        duplicated_values = list(values_count[values_count > 1].index)

        # go back into lower_values to find index of duplicata and their original value
        all_to_group = []
        for duplicata in duplicated_values:
            to_group = unique_values[clean_unique_values == duplicata]
            all_to_group.append(tuple(to_group))

        return all_to_group


    def get_groupings_acronym(self, col:pd.Series) -> list:
        """
        Gets groupings of values contained un "col" that are identified as being acronyms for one another. E.g. "hs" and "high school"

        :param col: pd.Series. A single column (feature). Assumed to contain strings.
        :return: list. Each element is a tuple of two strings. Each tuple represents the name of columns that could be grouped.
        """
        value = self.get_unique_values(col)
        value_lower = self.get_cleaned_values(value)
        df_info = pd.DataFrame({"value": value, "value_lower": value_lower})
        df_info["value_initial"] = df_info["value"].apply(lambda x: ''.join(i[0] for i in x.split()))
        df_info["value_initial_lower"] = df_info["value_lower"].apply(lambda x: ''.join(i[0] for i in x.split()))


        # find initials that could only come from one value
        init_lower_count = df_info["value_initial_lower"].value_counts()
        init_not_shared = list(init_lower_count[init_lower_count == 1].index)
        df_not_shared = df_info[df_info["value_initial_lower"].isin(init_not_shared)]

        # look into the values if the not shared intials are in it
        to_group = []
        for index, not_shared in df_not_shared.iterrows():
            for index_2, row in df_info.iterrows():
                if not_shared["value_initial_lower"] == row["value_lower"]:
                    to_group.append((row["value"], not_shared["value"]))

        return to_group


    def get_groupings_first_letter_of_another_value(self, col: pd.Series) -> list:
        """
        Gets groupings of values contained un "col" that are identified as being the first letter of another. E.g. "f" and "female"

        :param col: pd.Series. A single column (feature). Assumed to contain strings.
        :return: list. Each element is a tuple of two strings. Each tuple represents the name of columns that could be grouped.
        """
        values = self.get_unique_values(col)
        unique_clean_values = list(self.get_cleaned_values(values))

        # find values that are a single letter
        just_one_letter = []
        for value in unique_clean_values:
            if len(value) == 1:
                just_one_letter.append(value)

        starts_with = {k: [] for k in just_one_letter}

        # find the other values that start with those single letter
        for letter in just_one_letter:
            for value in unique_clean_values:
                if value[0] == letter and value != letter:
                    starts_with[letter].append(value)

        # if there is only one other value that starts with that letter, clearly someone was lazy and simply type "m" instead of "male"
        # otherwise (more than one value with this start letter) we cannot conclude
        to_group = []
        for letter in starts_with.keys():
            if len(starts_with[letter]) == 1:
                to_group.append((letter, starts_with[letter][0]))

        return to_group


    def get_levenshtein_close_values(self, col: pd.Series, threshold=85) -> list:
        """
        Gets groupings of values contained un "col" that are identified as being close to one another. E.g. "Montreal" and "Montrea"

        :param col: pd.Series. A single column (feature). Assumed to contain strings.
        :param threshold: int. Must be between 0 and 100. Threshold from which it is decided if two values are "similar" or not.
        :return: list. Each element is a tuple of two strings. Each tuple represents the name of columns that could be grouped.
        """
        values = self.get_unique_values(col)

        pairs_of_values = list(itertools.combinations(values, 2))
        high_score_pairs = []
        for pair in pairs_of_values:
            # .lower() could be replace by any pre-processing. Ideally, a copy of the pre-processed and original values should be kept everywhere
            # would allow to have access to pre-processed values for comparison, WHILE always knowing it was associated with what original value
            # Hence, can always know when something similar is found what original value it maps to.
            this_score = fuzz.ratio(pair[0].lower(), pair[1].lower())
            if this_score >= threshold:
                high_score_pairs.append(pair)
        return high_score_pairs


    # to apply only on values with more than one word
    # not optimized, but we can always repeat it
    # other way of doing levenhstein distance comparaison could be:
    # 1. you just compare with values "after" you in the list
    # 2. once you are tagged as "similar to someone", you are not comparing yourself afterwards (you are known and have been attributed to a group, that's it, it's over for you)
    def get_partial_levenshtein_close_values(self, col: pd.Series, threshold: int=85) -> list:
        """
        Gets groupings of values contained un "col" that are identified as being close to one another, with partial Levenshtein metric.
        E.g. "Toronto Raptors" and "Raptors". Any string fully contained in another string will be flagged as 100 similar with partial levenshtein metric.

        :param col: pd.Series. A single column (feature). Assumed to contain strings.
        :param threshold: int. Must be between 0 and 100. Threshold from which it is decided if two values are "similar" or not.
        :return: list. Each element is a tuple of two strings. Each tuple represents the name of columns that could be grouped.
        """
        values = self.get_unique_values(col)
        values_with_more_than_one_letter = values[values.str.len() > 1]
        values_with_more_than_one_word = values[values.str.contains(r" ")]

        # making sure that partial is only applied when multiple words
        # otherwise, something like "female" and "male" would get a score of 100 and be merged
        if values_with_more_than_one_word.empty:
            return []

        pairs_of_values = list(itertools.combinations(values_with_more_than_one_letter, 2))
        high_score_pairs = []
        for pair in pairs_of_values:

            this_score = fuzz.partial_ratio(pair[0].lower(), pair[1].lower())

            if this_score >= threshold:
                high_score_pairs.append(pair)

        return high_score_pairs


    def group_values_max_length(self, col: pd.Series, values_to_group: list=[] ) -> pd.Series:
        """
        Groups values passed into a single new value. E.g., if values_to_group is ["f", "female"], all "f" in "col" will be replaced by "female".
        Chooses the longest value in "values_to_group" as the value for all values mentioned in "values_to_group"

        :param col: pd.Series. A single column (feature). Assumed to contain strings.
        :param values_to_group: list. Each element is a tuple, containing values that needs to be grouped under the same value.
        :return: pd.Series. Column where all values mentioned in "values_to_group" now have the same value.
        """
        new_col = copy.deepcopy(col)
        if values_to_group:
            # let's use the value with the most letters as the umbrella name (heuristic for the moment, intent is to avoid losing too much info)
            # (deciding that new name in a smart way is a problem to solve on
            for group in values_to_group:
                new_name = max(group, key=len).lower()
                new_col.loc[col.isin(group)] = new_name
                print("values " + str(group) + "have been grouped under the value \'"+ str(new_name) + "\' for column '"+ str(col.name) + "'")
        return new_col


    def group_values_min_length(self, col: pd.Series, values_to_group: list) -> pd.Series:
        """
        Currently not used by any other method from the class.
        Groups values passed into a single new value. E.g., if values_to_group is ["f", "female"], all "f" in "col" will be replaced by "f".
        Chooses the smallest value in "values_to_group" as the value for all values mentioned in "values_to_group"

        :param col: pd.Series. A single column (feature). Assumed to contain strings.
        :param values_to_group: list. Each element is a tuple, containing values that needs to be grouped under the same value.
        :return: pd.Series. Column where all values mentioned in "values_to_group" now have the same value.
        """
        if values_to_group:
            # let's use the value with the most letters as the umbrella name (heuristic for the moment, intent is to avoid losing too much info)
            # (deciding that new name in a smart way is a problem to solve on
            new_col = copy.deepcopy(col)
            for group in values_to_group:
                new_name = min(group, key=len).lower()
                new_col.loc[col.isin(group)] = new_name
            return new_col


    def group_values_max_frequency(self, col: pd.Series, values_to_group: list) -> pd.Series:
        #TODO, most likely how tableau prep chooses their new name
        """
        Currently not used by any other method from the class.
        Groups values passed into a single new value. E.g., if "col" is ["f", "female", "female", "male"].
        and "values_to_group" is [("f", "female")], all "f" in "col" will be replaced by "female."
        Chooses the most frequent value in "values_to_group" as the value for all values mentioned in "values_to_group".
        """
        return None


    def get_cleaned_column(self, col:pd.Series) -> pd.Series:
        """
        Groups all values that are similar in "col" into a single value. Applies many different rules to identify similar values.

        :param col: pd.Series. A single column (feature). Assumed to contain strings.
        :return: pd.Series. Identical to "col", to the exception that all similar values as been grouped in a single value.
        """
        # basic
        basic_groupings = self.get_basic_groupings(col)
        col = self.group_values_max_length(col, basic_groupings)

        # levenshtein complete string
        values = col.unique()
        condition = True
        while condition:
            leven_groupings = self.get_levenshtein_close_values(col)
            col = self.group_values_max_length(col, leven_groupings)
            new_values = col.unique()

            if set(new_values) == set(values):
                condition = False
            else:
                values = new_values

        # levenshtein partial/substrings
        values = col.unique()
        condition = True
        while condition:
            leven_partial_groupings = self.get_partial_levenshtein_close_values(col)
            col = self.group_values_max_length(col, leven_partial_groupings)
            new_values = col.unique()

            if set(new_values) == set(values):
                condition = False
            else:
                values = new_values

        # acronym groupings
        acronym_groupings = self.get_groupings_acronym(col)
        col = self.group_values_max_length(col, acronym_groupings)

        # edge case of acronym
        # TO APPLY LAST BECAUSE MUST HAVE ONLY ONE STARTING WITH THIS LETTER
        # will group "f" and "female" if no other value starts with "f" other than "female"
        first_letter_groupings = self.get_groupings_first_letter_of_another_value(col)
        col = self.group_values_max_length(col, first_letter_groupings)

        return col


    def get_cleaned_df(self) -> pd.DataFrame:
        """
        For each column identified as categorical with strings, will group similar values under a single value.
        Uses self.df as the original DataFrame that needs cleaning.

        :return: pd.DataFrame. A DataFrame based on self.df where similar values are be grouped under a single value.
        """
        new_df = copy.deepcopy(self.df)
        features = self.get_string_categorical_features()
        for feat in features:
            new_df[feat] = self.get_cleaned_column(self.df[feat])

        return new_df
